import typing
from typing import Literal

from padmy.db import Database
from .network import convert_db
from ..logs import logs

try:
    import dash_cytoscape as cyto
except ImportError:
    logs.error('Both dash and dash_cytoscape need to be installed')
    raise

from dash import Dash, Output, Input, html
import json
import networkx as nx

cyto.load_extra_layouts()


class NodeData(typing.TypedDict):
    id: str
    count: int
    label: str


class EdgeData(typing.TypedDict):
    id: str
    source: str
    target: str


class DirectedElement(typing.TypedDict):
    data: NodeData | EdgeData


def get_directed_edges(g: nx.DiGraph) -> list[DirectedElement]:
    directed_edges: list[DirectedElement] = [
        {'data': {
            'id': f'{src}{tgt}',
            'source': src,
            'target': tgt
        }
        }
        for src, tgt in g.edges
    ]
    node_elems: list[DirectedElement] = [{'data': {'id': k, **v}} for k, v in g.nodes.data()]  # type: ignore
    return node_elems + directed_edges


_CIRCULAR_NODE_STYLE = {
    'source-arrow-color': 'red',
    'source-arrow-shape': 'triangle',
    'line-color': 'red'
}

_CIRCULAR_EDGE_STYLE = {
    'source-arrow-color': 'red',
    'source-arrow-shape': 'triangle',
    'line-color': 'red'
}


def get_cycles_styles(g: nx.DiGraph):
    styles = []
    for cycles in nx.simple_cycles(g):
        # Self referencing
        if len(cycles) == 1:
            cycle = cycles[0]
            styles += [
                {
                    'selector': f'#{cycle}',
                    'style': {**_CIRCULAR_NODE_STYLE, 'background-color': 'red'}
                },
                {
                    'selector': f'#{cycle}{cycle}',
                    'style': {**_CIRCULAR_EDGE_STYLE, 'background-color': 'red'}
                }
            ]
        else:
            _src, _tgt = cycles
            styles += [
                {
                    'selector': f'#{_src}',
                    'style': {**_CIRCULAR_NODE_STYLE, 'background-color': 'red'}
                },
                {
                    'selector': f'#{_tgt}',
                    'style': {**_CIRCULAR_NODE_STYLE, 'background-color': 'red'}
                },
                {
                    'selector': f'#{_src}{_tgt}',
                    'style': _CIRCULAR_EDGE_STYLE
                },
                {
                    'selector': f'#{_tgt}{_src}',
                    'style': _CIRCULAR_EDGE_STYLE
                }
            ]
    return styles


# See https://dash.plotly.com/cytoscape/layout for a full list
Layout = Literal['cose', 'breadthfirst', 'circle',
                 'preset', 'random', 'grid', 'concentric']
# needs cyto.load_extra_layouts() to be loaded
ExternalLayout = Literal['cose-bilkent', 'cola', 'euler', 'spread', 'dagre', 'klay']


def get_layout(g: nx.DiGraph,
               *,
               layout: Layout | ExternalLayout = 'klay',
               # Eg: {'width': '100%', 'height': '400px'}
               style: dict | None = None) -> cyto.Cytoscape:
    elements = get_directed_edges(g)
    _style = style or {'width': '100%', 'height': '800px'}
    stylesheet = [
        {
            'selector': 'node',
            'style': {
                'label': 'data(label)'
            }
        },
        {
            'selector': 'edge',
            'style': {
                'curve-style': 'bezier',
                'source-arrow-shape': 'triangle',
            }
        },
        *get_cycles_styles(g)
    ]
    return cyto.Cytoscape(
        layout={'name': layout},
        style=_style,
        elements=elements,
        stylesheet=stylesheet
    )


def run_simple_app(db: Database, port: int = 5555):
    g = convert_db(db)
    cyto_layout = get_layout(g)
    layout_id = "graph-layout"
    cyto_layout.id = layout_id  # type: ignore

    app = Dash(__name__)

    app.layout = html.Div([
        cyto_layout,
        html.Pre(id='cytoscape-tapNodeData-output'),
        html.Pre(id='cytoscape-tapEdgeData-output'),
    ])

    @app.callback(Output('cytoscape-tapNodeData-output', 'children'),
                  Input(layout_id, 'tapNodeData'))
    def _on_press_node(data):
        return json.dumps(data, indent=2)

    @app.callback(Output('cytoscape-tapEdgeData-output', 'children'),
                  Input(layout_id, 'tapEdgeData'))
    def _on_press_edge(data):
        return json.dumps(data, indent=2)

    app.run(port=port)  # type: ignore



import contextlib
import logging
import tempfile
from math import floor
from typing import cast

import asyncpg

from ..config import Config
from ..db import Database, Table, FKConstraint
from ..logs import logs
from ..utils import (
    check_cmd, pg_restore, pg_dump, create_db, drop_db, exec_psql, iterate_pg,
    check_extension_exists, check_tmp_table_exists,
    insert_many
)


def copy_database(database: str,
                  target_db: str,
                  schemas: list[str],
                  drop_public: bool = False):
    """
    Dumps and recreate the schemas from __repr__`database` into `target_db`.
    You can optionally specify `drop_public` to drop the public schema
    before restoring and avoid ' ERROR:  schema "public" already exists'
    (useful if you have *public* specified in your schemas list)
    """
    for cmd in ['pg_dump', 'createdb', 'dropdb', 'pg_restore', 'psql']:
        check_cmd(cmd)

    with tempfile.NamedTemporaryFile(suffix='.dump') as tmp_file:
        pg_dump(schemas=schemas,
                dump_path=tmp_file.name,
                database=database,
                options=['-Fc', '--schema-only'])
        drop_db(target_db, if_exists=True)
        create_db(database=target_db)

        if 'public' in schemas or drop_public:
            exec_psql(target_db, 'DROP SCHEMA public;')

        pg_restore(dump_path=tmp_file.name,
                   database=target_db)


def get_insert_child_fk_data_query(table: Table, child_table: Table) -> str:
    """
    Gets the query
    """
    joins = []

    def _fk_join(tmp_name: str, fk: FKConstraint, fk_index: int) -> str:
        return f'inner join {tmp_name} _s{fk_index} on _s{fk_index}.{fk.column_name} = t.{fk.foreign_column_name}'

    for i, _fk in enumerate(child_table.foreign_keys):
        if _fk.foreign_full_name != table.full_name:
            continue
        joins.append(_fk_join(child_table.tmp_name, _fk, i))

    query = f"""
    INSERT INTO {table.tmp_name}
    SELECT * from {table.full_name} t
    """ + '\n'.join(joins) + '\n ON CONFLICT DO NOTHING'

    return query


def get_insert_data_query(table: Table):
    where_str = 'and '.join(f't2.{_pk.column_name} = t1.{_pk.column_name}' for _pk in table.primary_keys)
    query = f"""
    INSERT into {table.tmp_name}
    SELECT * from {table.full_name} t1
    where not exists(select null from {table.tmp_name} t2 where {where_str})
    limit $1
    """
    return query


async def _insert_leaf_table(conn: asyncpg.Connection, table: Table, table_size: int):
    query = f"CREATE TEMP TABLE {table.tmp_name} ON COMMIT DROP AS SELECT * from {table.full_name}"
    args = []
    if table_size > 0:
        query = f"{query} TABLESAMPLE SYSTEM_ROWS($1)"
        args.append(table_size)
    logs.debug(f'{query} {args}')
    await conn.execute(query, *args)


async def _insert_node_table(conn: asyncpg.Connection,
                             table: Table,
                             table_size: int):
    # Creating the table
    query = f"CREATE TEMP TABLE {table.tmp_name} (LIKE {table.full_name} INCLUDING ALL)  ON COMMIT DROP"
    logs.debug(query)
    await conn.execute(query)

    # Inserting data from child table
    for _child_table in table.child_tables:
        table_exists = await check_tmp_table_exists(conn, _child_table.full_name)
        if not table_exists:
            continue
        query = get_insert_child_fk_data_query(table, _child_table)
        logs.debug(query)
        await conn.execute(query)

    count = cast(int | None, await conn.fetchval(f'SELECT count(*) from {table.tmp_name}'))
    if count is None:
        raise NotImplementedError('Got empty table')

    match count:
        case count if count == table_size:
            pass
        case count if count < table_size:
            query = get_insert_data_query(table)
            _limit = table_size - count
            logs.debug(f'{query} limit: {_limit}')
            await conn.execute(query, _limit)
        case count if count > table_size:
            logs.warning(f'Sample size cannot be reached (got {count}, expected {table_size})')
        case _:
            raise NotImplementedError(f'Got invalid count: {count} (table_size: {table_size})')


async def process_table(table: Table, conn: asyncpg.Connection) -> set[Table]:
    if table.has_been_processed:
        raise ValueError(f'table {table.full_name!r} has already been processed')

    is_leaf = not table.has_children

    if table.sample_size is None:
        raise ValueError(f'Got empty sample_size for {table.full_name!r}')

    table_size = floor(table.count * table.sample_size / 100)
    if is_leaf:
        logs.info(f'\t Inserting {table.full_name}')
        await _insert_leaf_table(conn, table, table_size)
    else:
        if table.children_has_been_processed:
            logs.info(f'\t Inserting {table.full_name}')
            await _insert_node_table(conn, table, table_size)
        else:
            return {_child_table for _child_table in table.child_tables_safe
                    if not _child_table.has_been_processed
                    }

    if logs.level == logging.DEBUG:
        c = await conn.fetchval(f'SELECT count(*) from {table.tmp_name}')
        logs.debug(f'{table.tmp_name}: {c}')

    table.has_been_processed = True

    return {_parent_table for _parent_table
            in table.parent_tables_safe
            if not _parent_table.has_been_processed}


@contextlib.asynccontextmanager
async def disable_trigger(conn: asyncpg.Connection):
    await conn.execute("SET session_replication_role = 'replica'")
    try:
        yield
    finally:
        await conn.execute("SET session_replication_role = 'origin'")


async def create_temp_tables(conn: asyncpg.Connection,
                             tables: list[Table]):
    """
    Creates temporary tables with a sample of the original table
    """

    # We start from the leaves
    _tables = set(table for table in tables if not table.has_children)

    while len(_tables) > 0:
        _parent_tables = set()

        for table in _tables:
            if table.has_been_processed:
                continue

            _parent_tables = _parent_tables | await process_table(table, conn)

        # To avoid infinite loop
        if _tables == _parent_tables:
            _not_processed = [_table for _table in tables if not _table.children_has_been_processed]

            for _table in _not_processed:
                logs.error(_table.full_name)
                for c in _table.child_tables:
                    logs.error(f'\t {c.full_name} ({c.has_been_processed})')
            raise ValueError('Cyclic foreign keys detected. '
                             f'Possible tables are: {", ".join(x.full_name for x in _not_processed)}. '
                             'Run `analyze` with `--show-graphs` to debug.')

        _tables = _parent_tables


async def sample_database(conn: asyncpg.Connection,
                          target_conn: asyncpg.Connection,
                          config: Config,
                          db: Database,
                          *,
                          chunk_size: int = 1_000):
    """
    From top table to bottom
    """

    # getting the leaves
    has_ext = await check_extension_exists(conn, 'tsm_system_rows')
    if not has_ext:
        await conn.execute('CREATE EXTENSION tsm_system_rows')

    async with conn.transaction():
        await create_temp_tables(conn, db.tables)

        # Safety check
        _not_processed_tables = [x.full_name for x in db.tables if not x.has_been_processed]
        if _not_processed_tables:
            raise NotImplementedError(f'Found {len(_not_processed_tables)} tables that has not been'
                                      f'processed: {",".join(_not_processed_tables)}')
        #
        if logs.level == logging.DEBUG:
            for table in db.tables:
                c = await conn.fetchval(f'SELECT count(*) from {table.tmp_name}')
                logs.debug(f'{table.tmp_name}: {c}')

        logs.info('Done creating temporary tables, inserting to new database')

        async with disable_trigger(target_conn):
            for table in db.tables:
                logs.info(f'Inserting to {table.full_name}')
                query = f'SELECT * from {table.tmp_name}'
                async for chunk in iterate_pg(conn, query, chunk_size=chunk_size):
                    await insert_many(target_conn, table.full_name, chunk)



import dataclasses
import datetime as dt
import difflib
import filecmp
import functools
from pathlib import Path

from asyncpg import Connection
from asyncpg.exceptions import UndefinedTableError

from padmy.logs import logs
from padmy.utils import exec_file, pg_dump, exec_psql_file, insert_many
from .utils import get_files, iter_migration_files, MigrationFile

_GET_LATEST_COMMIT_QUERY = """
SELECT file_ts, file_name
FROM public.migration
ORDER BY applied_at DESC, file_ts DESC
LIMIT 1
"""

# PG_GLOBAL_DUMP_DIR = GLOBAL_TMP / 'matching-pg-backup' / 'dump'
SETUP_FILE = Path(__file__).parent / 'db.sql'


class MigrationError(Exception):
    def __init__(self, msg: str, diff: str):
        self.msg = msg
        self.diff = diff


async def migrate_setup(conn: Connection):
    logs.info('Setting up migration tables')
    await exec_file(conn, SETUP_FILE)
    logs.info('Done')


def _verify_migration(
        database: str,
        schemas: list[str],
        migration_id: str,
        up_file: Path,
        down_file: Path,
        dump_dir: Path,
):
    _before_dump, _after_dump = f'{migration_id}-before.sql', f'{migration_id}-after.sql'

    # Dump before
    pg_dump(database, schemas, dump_path=str(dump_dir / _before_dump),
            options=['-E', 'utf8', '--schema-only'])

    logs.debug(f'Applying {up_file.name}')
    exec_psql_file(database, str(up_file))

    logs.debug(f'Applying {down_file.name}')
    exec_psql_file(database, str(down_file))

    # Dump after
    pg_dump(database, schemas, dump_path=str(dump_dir / _after_dump),
            options=['-E', 'utf8', '--schema-only'])

    _before_dump_file, _after_dump_file = dump_dir / _before_dump, dump_dir / _after_dump

    _no_diff = filecmp.cmp(_before_dump_file, _after_dump_file)

    if not _no_diff:
        diff = difflib.unified_diff(_before_dump_file.read_text().split('\n'), _after_dump_file.read_text().split('\n'),
                                    fromfile=_before_dump_file.name,
                                    tofile=_after_dump_file.name)
        raise MigrationError(f'Difference found for migration: {migration_id}',
                             diff='\n'.join(diff))


def migrate_verify(database: str,
                   schemas: list[str],
                   dump_dir: Path,
                   migration_folder: Path,
                   ):
    """
    Verifies that the up/down migration is correct
    """
    _pg_dump = functools.partial(pg_dump,
                                 database=database,
                                 schemas=schemas)

    migration_files = get_files(migration_folder)
    logs.info(f'Found {len(migration_files)} migration files')

    logs.info('Checking...')
    _down_files: list[MigrationFile] = []
    try:
        for i, (_up_file, _down_file) in enumerate(iter_migration_files(migration_files)):
            _verify_migration(database=database,
                              schemas=schemas,
                              migration_id=_up_file.file_id,
                              up_file=_up_file.path,
                              down_file=_down_file.path,
                              dump_dir=dump_dir)
            # Need to be reapplied for the next tests
            exec_psql_file(database, str(_up_file.path))
            _down_files.append(_down_file)
    except MigrationError as e:
        logs.error(e.msg)
        logs.error(e.diff)
        raise e

    # Restoring to initial state
    logs.info('Restoring to initial state')
    for _down_file in reversed(_down_files):
        exec_psql_file(database, str(_down_file.path))
    # else:
    #     if last_down_file:
    #         exec_psql_file(database, str(last_down_file.path))

    logs.info('Everything ok!')


@dataclasses.dataclass
class LatestMigration:
    timestamp: dt.datetime
    file_name: str

    @classmethod
    def load(cls, data: dict):
        return cls(
            timestamp=data['file_ts'],
            file_name=data['file_name']
        )


class NoSetupTableError(Exception):
    pass


async def _get_latest_migration(conn: Connection):
    try:
        resp = await conn.fetchrow(_GET_LATEST_COMMIT_QUERY)
    except UndefinedTableError as e:
        if 'relation "public.migration" does not exist' in str(e):
            raise NoSetupTableError(
                'Could not find table table "public.migration", did you forget to setup the table'
                ' by running "migration setup" ?'
            )
        else:
            raise
    return LatestMigration.load(dict(resp)) if resp else None


async def migrate_up(conn: Connection,
                     folder: Path,
                     *,
                     nb_migrations: int = -1,
                     ):
    """ Migrate from the latest migration applied available in the database
    to the latest one in the `folder` dir.
    """
    latest_migration = await _get_latest_migration(conn)
    logs.info(f'Latest timestamp: {latest_migration.timestamp}'
              if latest_migration
              else 'No previous migration found')
    migration_files = get_files(folder)

    migrations_to_apply = [
        _up_file for _up_file, _ in iter_migration_files(migration_files)
        if latest_migration is None or (
                _up_file.ts >= latest_migration.timestamp and
                _up_file.path.name != latest_migration.file_name
        )
    ]
    if nb_migrations > 0:
        migrations_to_apply = migrations_to_apply[:nb_migrations]

    if not migrations_to_apply:
        logs.info('No migrations to apply')
        return

    logs.info(f'Found {len(migrations_to_apply)} migrations to apply')
    async with conn.transaction():
        for _migration in migrations_to_apply:
            logs.info(f'Running {_migration.path.name}...')
            try:
                await exec_file(conn, _migration.path)
            except Exception as e:
                logs.error(f'Failed to execute migration_id "{_migration.file_id}"')
                raise e
            commit_data = {
                'file_ts': _migration.ts,
                'file_id': _migration.file_id,
                'migration_type': 'up',
                'file_name': _migration.path.name
            }
            await insert_many(conn, 'public.migration', [commit_data])

    logs.info('Done!')


async def migrate_down(conn: Connection,
                       folder: Path,
                       *,
                       nb_migrations: int = -1,
                       # migration_id: str = None
                       ):
    """
    Rollback `nb_migrations` back (default -1 for all available)
    """
    latest_migration = await _get_latest_migration(conn)

    migration_files = get_files(folder, reverse=True)

    rollback_to_apply = [
        _down_file for _, _down_file in iter_migration_files(migration_files)
        if latest_migration is None or (
                _down_file.ts <= latest_migration.timestamp and
                _down_file.path.name != latest_migration.file_name
        )
    ]

    if nb_migrations > 0:
        rollback_to_apply = rollback_to_apply[:nb_migrations]

    nb_files = len(rollback_to_apply)
    if nb_files == 0:
        logs.info('No rollback files to apply')
        return

    logs.info(f'Found {nb_files} rollback files to apply')
    async with conn.transaction():
        for _rollback in rollback_to_apply:
            logs.info(f'Running {_rollback.path.name}...')
            try:
                await exec_file(conn, _rollback.path)
            except Exception as e:
                logs.error(f'Failed to execute migration_id "{_rollback.file_id}"')
                raise e

            commit_data = {
                'file_ts': _rollback.ts,
                'file_id': _rollback.file_id,
                'migration_type': 'down',
                'file_name': _rollback.path.name
            }
            await insert_many(conn, 'public.migration', [commit_data])

    logs.info('Done!')



import os

PG_DATABASE = os.getenv('PG_DATABASE', 'postgres')
PG_HOST = os.getenv('PG_HOST', 'localhost')
PG_PORT = int(os.getenv('PG_PORT', '5432'))
PG_USER = os.getenv('PG_USER', 'postgres')
PG_PASSWORD = os.getenv('PG_PASSWORD', 'postgres')

# Migration
SQL_DIR = os.getenv('SQL_DIR')
MIGRATION_DIR = os.getenv('MIGRATION_DIR')



import subprocess
from importlib import reload
from pathlib import Path
from typing import Sequence, AsyncIterator, Callable, TypeVar, cast

import asyncpg
from piou import Option, Derived, Password

from padmy import env
from .env import PG_HOST, PG_PORT, PG_USER, PG_DATABASE, PG_PASSWORD
from .logs import logs

PgHost = Option(PG_HOST, '--host', help='PG Host')
PgPort = Option(PG_PORT, '--port', help='PG Port')
PgUser = Option(PG_USER, '--user', help='PG User')
PgPassword = Option(PG_PASSWORD, '-p', help='PG Password')
PgDatabase = Option(PG_DATABASE, '--db', help='PG Database')


def get_pg_root(
        pg_host: str = PgHost,
        pg_port: int = PgPort,
        pg_user: str = PgUser,
        pg_password: Password = PgPassword,
):
    return f'postgresql://{pg_user}:{pg_password}@{pg_host}:{pg_port}'


def get_pg_url(
        pg_url: str = Derived(get_pg_root),
        pg_database: str = PgDatabase
):
    return f'{pg_url}/{pg_database}'


async def get_pg(
        pg_url: str = Derived(get_pg_url)
):
    return await asyncpg.connect(pg_url)


def exec_cmd(cmd: Sequence[str | int],
             env: dict | None = None,
             *,
             on_stderr: Callable | None = None) -> str:
    _cmd = cmd if isinstance(cmd, str) else ' '.join(str(x) for x in cmd)
    logs.debug(f'Executing cmd: {_cmd}')
    stdout, stderr = subprocess.Popen(_cmd,
                                      shell=True,
                                      stdout=subprocess.PIPE,
                                      stderr=subprocess.PIPE,
                                      env=env).communicate()
    if stderr:
        _msg = stderr.decode('utf-8')
        logs.warning(_msg)
        if on_stderr:
            on_stderr(_msg)
    return stdout.decode('utf-8')


def has_cmd(cmd: str) -> bool:
    return exec_cmd(f'command -v "{cmd}"') != ''


class CommandNotFound(Exception):
    def __init__(self, cmd: str):
        super().__init__()
        self.cmd = cmd


def check_cmd(cmd: str):
    if not has_cmd(cmd):
        raise CommandNotFound(cmd=cmd)


def get_pg_envs():
    reload(env)
    return {'PGPASSWORD': env.PG_PASSWORD,
            'PGUSER': env.PG_USER,
            'PGHOST': env.PG_HOST,
            'PGPORT': str(env.PG_PORT)}


def _on_pg_error(msg: str):
    if 'error' in msg:
        raise ValueError(msg)


def pg_dump(
        database: str,
        schemas: list[str],
        dump_path: str,
        options: list[str] | None = None
):
    _schemas = '|'.join(x for x in schemas)
    cmd = [
        'pg_dump',
        '-n', f"'({_schemas})'",
        '-d', database
    ]

    if options:
        cmd += options

    cmd += ['>', dump_path]

    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


def pg_restore(
        database: str,
        dump_path: str,
):
    cmd = [
        'pg_restore', '-d', database,
        dump_path
    ]
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


def create_db(database: str):
    cmd = ['createdb', database]
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


def drop_db(database: str, if_exists: bool = True):
    cmd = ['dropdb']
    if if_exists:
        cmd.append('--if-exists')
    cmd.append(database)
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


def exec_psql_file(database: str, sql_file: str):
    cmd = ['psql', '-f', sql_file, '-d', database]
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


def exec_psql(database: str, query: str):
    cmd = ['psql', '-c', f"'{query}'", '-d', database]
    exec_cmd(cmd, env=get_pg_envs(), on_stderr=_on_pg_error)


async def exec_file(conn: asyncpg.Connection, file: Path):
    await conn.execute(file.read_text())


async def iterate_pg(conn: asyncpg.Connection,
                     query: str,
                     *args,
                     from_offset: int = 0,
                     chunk_size: int = 500,
                     timeout: int | None = None) -> AsyncIterator[list[asyncpg.Record]]:
    async with conn.transaction():
        cur: asyncpg.connection.cursor.Cursor = await conn.cursor(query, *args)
        if from_offset:
            await cur.forward(from_offset, timeout=timeout)
        while data := await cur.fetch(chunk_size, timeout=timeout):
            yield cast(list[asyncpg.Record], data)


_EXT_EXISTS_QUERY = """
SELECT EXISTS(
    SELECT FROM pg_extension WHERE extname = $1
)
"""


async def check_extension_exists(conn: asyncpg.Connection,
                                 extension: str) -> bool:
    _exists = await conn.fetchval(_EXT_EXISTS_QUERY, extension)
    if _exists is None:
        raise NotImplementedError()
    return _exists


_TMP_TABLE_EXISTS_QUERY = """
    SELECT EXISTS (
        SELECT 
        FROM   information_schema.tables 
        WHERE  table_schema LIKE 'pg_temp_%'
        AND table_name = $1
    )
"""


async def check_tmp_table_exists(conn: asyncpg.Connection, table: str) -> bool:
    exists = await conn.fetchval(_TMP_TABLE_EXISTS_QUERY, table)
    if exists is None:
        raise NotImplementedError()
    return exists


async def insert_many(conn: asyncpg.Connection,
                      table: str,
                      data: list[dict]):
    """
    Keys in dict must have the same order
    """
    keys = tuple(data[0].keys())
    fields = ', '.join(f'"{k}"' for k in keys)
    values = ', '.join(f'${i + 1}' for i, _ in enumerate(keys))
    query = f'INSERT INTO {table} ({fields}) VALUES ({values})'
    _data = [tuple(x.values()) for x in data]
    await conn.executemany(query, _data)


X = TypeVar('X')


def get_first(*it: X, fn: Callable) -> X | None:
    for v in it:
        if fn(v):
            return v
    return None


async def get_conn(pool: asyncpg.Pool, fn: Callable):
    async with pool.acquire() as conn:
        await fn(conn=conn)



import os
import tempfile
from pathlib import Path

from asyncpg import Connection
from piou import Option, Derived, CommandGroup, Password

from padmy.env import SQL_DIR, MIGRATION_DIR
from padmy.logs import logs
from padmy.utils import get_pg, PgHost, PgPort, PgUser, PgDatabase, PgPassword

MIGRATION_DESCRIPTION = """
Utilities to handle schema migrations with PostgresSQL.  
If it's the first time running theses scripts please run 
**`poetry run cli migrate setup`** to setup the database's table
"""

migration = CommandGroup('migrate',
                         help='Migration utilities',
                         description=MIGRATION_DESCRIPTION)

MigrationDir = Option(MIGRATION_DIR or ..., '--sql-dir', help='Directory containing the migration files')
SQLDir = Option(SQL_DIR or ..., '--sql-dir', help='Directory containing the table definitions')


@migration.command(cmd='new-sql')
def new_sql_file(
        sql_dir: Path = SQLDir,
        position: int = Option(1, help='Position where to insert the sql file'),

):
    """
    Utility to insert a new sql file between other sql files.
    """
    from .new_sql import create_sql_file

    create_sql_file(sql_dir, position)


@migration.command(cmd='apply-sql')
async def apply_sql_files(
        pg_conn: Connection = Derived(get_pg),
        sql_dir: Path = SQLDir,
):
    """
    Utility to insert a new sql file between other sql files.
    """
    from ..utils import exec_file
    for file in sorted(sql_dir.glob('*.sql')):
        logs.info(f'Applying {file.name!r}')
        await exec_file(pg_conn, file)




@migration.command(cmd='new', help='Creates 2 new files for a migration (up and down)')
def new_migrate_file(migration_folder: Path = MigrationDir):
    """
    Creates 2 new files for a migration (up and down).
    When applying *up.sql* then *down.sql*, the database must be
    in the state it was previously.
    To check if your **down** file works as expected, you can use the `verify` command.
    """
    from .create_files import create_new_migration
    create_new_migration(migration_folder)


@migration.command(cmd='up', help='Migrate database to the new schema')
async def migrate_up_main(
        pg_conn: Connection = Derived(get_pg),
        sql_dir: Path = MigrationDir,
        nb_migrations: int = Option(1, '-n', '--nb-migrations', help='Number of migrations to apply')
):
    from .migration import migrate_up, NoSetupTableError

    try:
        await migrate_up(pg_conn, folder=sql_dir,
                         nb_migrations=nb_migrations)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd='down', help='Rollback the database to the specified migration')
async def migrate_down_main(
        conn: Connection = Derived(get_pg),
        sql_dir: Path = MigrationDir,
        nb_rollbacks: int = Option(1, '-n', '--nb-rollbacks', help='Number of rollbacks to apply')
):
    from .migration import migrate_down, NoSetupTableError
    try:
        await migrate_down(conn, folder=sql_dir,
                           nb_migrations=nb_rollbacks)
    except NoSetupTableError as e:
        logs.critical(e, exc_info=False)


@migration.command(cmd='setup', help='Create the tables that will contain the migration metadata')
async def migrate_setup_main(
        pg_conn: Connection = Derived(get_pg)
):
    from .migration import migrate_setup
    await migrate_setup(pg_conn)


@migration.command(cmd='verify', help='Verify that a migration is valid')
def migrate_verify_main(
        db: str = PgDatabase,
        pg_host: str = PgHost,
        pg_port: int = PgPort,
        pg_user: str = PgUser,
        pg_password: Password = PgPassword,
        schemas: list[str] = Option(..., '--schemas', help='Schemas impacted by the migration'),
        sql_dir: Path = MigrationDir):

    os.environ['PG_HOST'] = pg_host
    os.environ['PG_PORT'] = str(pg_port)
    os.environ['PG_USER'] = pg_user
    os.environ['PG_PASSWORD'] = pg_password

    from .migration import migrate_verify, MigrationError

    try:
        with tempfile.TemporaryDirectory() as dump_dir:
            migrate_verify(database=db,
                           migration_folder=sql_dir,
                           schemas=schemas,
                           dump_dir=Path(dump_dir))
    except MigrationError as e:
        logs.error(e.msg)
        logs.debug(e.diff)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['padmy', 'padmy.anonymize', 'padmy.migration', 'padmy.sampling']

package_data = \
{'': ['*']}

install_requires = \
['Faker>=13.15.1,<14.0.0',
 'PyYAML>=6.0,<7.0',
 'asyncpg>=0.27.0,<0.28.0',
 'piou>=0.13.1,<0.14.0',
 'typing-extensions>=4.3.0,<5.0.0']

extras_require = \
{'network': ['networkx>=2.8.5,<3.0.0',
             'dash>=2.6.0,<3.0.0',
             'dash-cytoscape>=0.3.0,<0.4.0']}

entry_points = \
{'console_scripts': ['cli = run:run']}

setup_kwargs = {
    'name': 'padmy',
    'version': '0.4.0',
    'description': '',
    'long_description': 'None',
    'author': 'andarius',
    'author_email': 'julien.brayere@tracktor.fr',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
